@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 #
 # Copyright 2001 Google Inc. All Rights Reserved.
 #
@@ -19,23 +19,24 @@
 Projects that use ninja themselves should either write a similar script
 or use a meta-build system that supports Ninja output."""
 
-from __future__ import print_function
-
 from optparse import OptionParser
 import os
-import pipes
-import string
+import shlex
 import subprocess
 import sys
+from typing import Optional, Union, Dict, List, Any, TYPE_CHECKING
 
 sourcedir = os.path.dirname(os.path.realpath(__file__))
 sys.path.insert(0, os.path.join(sourcedir, 'misc'))
-import ninja_syntax
+if TYPE_CHECKING:
+    import misc.ninja_syntax as ninja_syntax
+else:
+    import ninja_syntax
 
 
 class Platform(object):
     """Represents a host/target platform and its specific build attributes."""
-    def __init__(self, platform):
+    def __init__(self, platform: Optional[str]) -> None:
         self._platform = platform
         if self._platform is not None:
             return
@@ -66,55 +67,55 @@ class Platform(object):
             self._platform = 'dragonfly'
 
     @staticmethod
-    def known_platforms():
+    def known_platforms() -> List[str]:
       return ['linux', 'darwin', 'freebsd', 'openbsd', 'solaris', 'sunos5',
               'mingw', 'msvc', 'gnukfreebsd', 'bitrig', 'netbsd', 'aix',
               'dragonfly']
 
-    def platform(self):
-        return self._platform
+    def platform(self) -> str:
+        return self._platform  # type: ignore # Incompatible return value type
 
-    def is_linux(self):
+    def is_linux(self) -> bool:
         return self._platform == 'linux'
 
-    def is_mingw(self):
+    def is_mingw(self) -> bool:
         return self._platform == 'mingw'
 
-    def is_msvc(self):
+    def is_msvc(self) -> bool:
         return self._platform == 'msvc'
 
-    def msvc_needs_fs(self):
-        popen = subprocess.Popen(['cl', '/nologo', '/?'],
+    def msvc_needs_fs(self) -> bool:
+        popen = subprocess.Popen(['cl', '/nologo', '/help'],
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE)
         out, err = popen.communicate()
         return b'/FS' in out
 
-    def is_windows(self):
+    def is_windows(self) -> bool:
         return self.is_mingw() or self.is_msvc()
 
-    def is_solaris(self):
+    def is_solaris(self) -> bool:
         return self._platform == 'solaris'
 
-    def is_aix(self):
+    def is_aix(self) -> bool:
         return self._platform == 'aix'
 
-    def is_os400_pase(self):
-        return self._platform == 'os400' or os.uname().sysname.startswith('OS400')
+    def is_os400_pase(self) -> bool:
+        return self._platform == 'os400' or os.uname().sysname.startswith('OS400')  # type: ignore # Module has no attribute "uname"
 
-    def uses_usr_local(self):
+    def uses_usr_local(self) -> bool:
         return self._platform in ('freebsd', 'openbsd', 'bitrig', 'dragonfly', 'netbsd')
 
-    def supports_ppoll(self):
+    def supports_ppoll(self) -> bool:
         return self._platform in ('freebsd', 'linux', 'openbsd', 'bitrig',
                                   'dragonfly')
 
-    def supports_ninja_browse(self):
+    def supports_ninja_browse(self) -> bool:
         return (not self.is_windows()
                 and not self.is_solaris()
                 and not self.is_aix())
 
-    def can_rebuild_in_place(self):
+    def can_rebuild_in_place(self) -> bool:
         return not (self.is_windows() or self.is_aix())
 
 class Bootstrap:
@@ -125,37 +126,43 @@ class Bootstrap:
     It also proxies all calls to an underlying ninja_syntax.Writer, to
     behave like non-bootstrap mode.
     """
-    def __init__(self, writer, verbose=False):
+    def __init__(self, writer: ninja_syntax.Writer, verbose: bool = False) -> None:
         self.writer = writer
         self.verbose = verbose
         # Map of variable name => expanded variable value.
-        self.vars = {}
+        self.vars: Dict[str, str] = {}
         # Map of rule name => dict of rule attributes.
-        self.rules = {
+        self.rules: Dict[str, Dict[str, Any]] = {
             'phony': {}
         }
 
-    def comment(self, text):
+    def comment(self, text: str) -> None:
         return self.writer.comment(text)
 
-    def newline(self):
+    def newline(self) -> None:
         return self.writer.newline()
 
-    def variable(self, key, val):
+    def variable(self, key: str, val: str) -> None:
         # In bootstrap mode, we have no ninja process to catch /showIncludes
         # output.
         self.vars[key] = self._expand(val).replace('/showIncludes', '')
         return self.writer.variable(key, val)
 
-    def rule(self, name, **kwargs):
+    def rule(self, name: str, **kwargs: Any) -> None:
         self.rules[name] = kwargs
         return self.writer.rule(name, **kwargs)
 
-    def build(self, outputs, rule, inputs=None, **kwargs):
+    def build(
+        self,
+        outputs: Union[str, List[str]],
+        rule: str,
+        inputs: Optional[Union[str, List[str]]] = None,
+        **kwargs: Any
+    ) -> List[str]:
         ruleattr = self.rules[rule]
         cmd = ruleattr.get('command')
         if cmd is None:  # A phony rule, for example.
-            return
+            return  # type: ignore # Return value expected
 
         # Implement just enough of Ninja variable expansion etc. to
         # make the bootstrap build work.
@@ -170,23 +177,23 @@ class Bootstrap:
 
         return self.writer.build(outputs, rule, inputs, **kwargs)
 
-    def default(self, paths):
+    def default(self, paths: Union[str, List[str]]) -> None:
         return self.writer.default(paths)
 
-    def _expand_paths(self, paths):
+    def _expand_paths(self, paths: Optional[Union[str, List[str]]]) -> str:
         """Expand $vars in an array of paths, e.g. from a 'build' block."""
         paths = ninja_syntax.as_list(paths)
         return ' '.join(map(self._shell_escape, (map(self._expand, paths))))
 
-    def _expand(self, str, local_vars={}):
+    def _expand(self, str: str, local_vars: Dict[str, str] = {}) -> str:
         """Expand $vars in a string."""
         return ninja_syntax.expand(str, self.vars, local_vars)
 
-    def _shell_escape(self, path):
+    def _shell_escape(self, path: str) -> str:
         """Quote paths containing spaces."""
         return '"%s"' % path if ' ' in path else path
 
-    def _run_command(self, cmdline):
+    def _run_command(self, cmdline: str) -> None:
         """Run a subcommand, quietly.  Prints the full command on error."""
         try:
             if self.verbose:
@@ -216,7 +223,10 @@ parser.add_option('--debug', action='store_true',
 parser.add_option('--profile', metavar='TYPE',
                   choices=profilers,
                   help='enable profiling (' + '/'.join(profilers) + ')',)
-parser.add_option('--with-gtest', metavar='PATH', help='ignored')
+parser.add_option('--gtest-source-dir', metavar='PATH',
+                  help='Path to GoogleTest source directory. If not provided ' +
+                       'GTEST_SOURCE_DIR will be probed in the environment. ' +
+                       'Tests will not be built without a value.')
 parser.add_option('--with-python', metavar='EXE',
                   help='use EXE as the Python interpreter',
                   default=os.path.basename(sys.executable))
@@ -236,7 +246,7 @@ else:
 
 BUILD_FILENAME = 'build.ninja'
 ninja_writer = ninja_syntax.Writer(open(BUILD_FILENAME, 'w'))
-n = ninja_writer
+n: Union[ninja_syntax.Writer, Bootstrap] = ninja_writer
 
 if options.bootstrap:
     # Make the build directory.
@@ -247,7 +257,7 @@ if options.bootstrap:
     # Wrap ninja_writer with the Bootstrapper, which also executes the
     # commands.
     print('bootstrapping ninja...')
-    n = Bootstrap(n, verbose=options.verbose)
+    n = Bootstrap(n, verbose=options.verbose)  # type: ignore # Incompatible types in assignment
 
 n.comment('This file is used to build ninja itself.')
 n.comment('It is generated by ' + os.path.basename(__file__) + '.')
@@ -264,7 +274,7 @@ n.variable('configure_args', ' '.join(configure_args))
 env_keys = set(['CXX', 'AR', 'CFLAGS', 'CXXFLAGS', 'LDFLAGS'])
 configure_env = dict((k, os.environ[k]) for k in os.environ if k in env_keys)
 if configure_env:
-    config_str = ' '.join([k + '=' + pipes.quote(configure_env[k])
+    config_str = ' '.join([k + '=' + shlex.quote(configure_env[k])
                            for k in configure_env])
     n.variable('configure_env', config_str + '$ ')
 n.newline()
@@ -275,17 +285,17 @@ if platform.is_msvc():
     CXX = 'cl'
     objext = '.obj'
 
-def src(filename):
+def src(filename: str) -> str:
     return os.path.join('$root', 'src', filename)
-def built(filename):
+def built(filename: str) -> str:
     return os.path.join('$builddir', filename)
-def doc(filename):
+def doc(filename: str) -> str:
     return os.path.join('$root', 'doc', filename)
-def cc(name, **kwargs):
+def cc(name: str, **kwargs: Any) -> List[str]:
     return n.build(built(name + objext), 'cxx', src(name + '.c'), **kwargs)
-def cxx(name, **kwargs):
+def cxx(name: str, **kwargs: Any) -> List[str]:
     return n.build(built(name + objext), 'cxx', src(name + '.cc'), **kwargs)
-def binary(name):
+def binary(name: str) -> str:
     if platform.is_windows():
         exe = name + '.exe'
         n.build(name, 'phony', exe)
@@ -305,14 +315,26 @@ if platform.is_msvc():
 else:
     n.variable('ar', configure_env.get('AR', 'ar'))
 
+def search_system_path(file_name: str) -> Optional[str]:  # type: ignore # Missing return statement
+  """Find a file in the system path."""
+  for dir in os.environ['path'].split(';'):
+    path = os.path.join(dir, file_name)
+    if os.path.exists(path):
+      return path
+
+# Note that build settings are separately specified in CMakeLists.txt and
+# these lists should be kept in sync.
 if platform.is_msvc():
+    if not search_system_path('cl.exe'):
+        raise Exception('cl.exe not found. Run again from the Developer Command Prompt for VS')
     cflags = ['/showIncludes',
               '/nologo',  # Don't print startup banner.
+              '/utf-8',
               '/Zi',  # Create pdb with debug info.
               '/W4',  # Highest warning level.
               '/WX',  # Warnings as errors.
               '/wd4530', '/wd4100', '/wd4706', '/wd4244',
-              '/wd4512', '/wd4800', '/wd4702', '/wd4819',
+              '/wd4512', '/wd4800', '/wd4702',
               # Disable warnings about constant conditional expressions.
               '/wd4127',
               # Disable warnings about passing "this" during initialization.
@@ -320,6 +342,7 @@ if platform.is_msvc():
               # Disable warnings about ignored typedef in DbgHelp.h
               '/wd4091',
               '/GR-',  # Disable RTTI.
+              '/Zc:__cplusplus',
               # Disable size_t -> int truncation warning.
               # We never have strings or arrays larger than 2**31.
               '/wd4267',
@@ -339,6 +362,7 @@ else:
               '-Wno-unused-parameter',
               '-fno-rtti',
               '-fno-exceptions',
+              '-std=c++14',
               '-fvisibility=hidden', '-pipe',
               '-DNINJA_PYTHON="%s"' % options.with_python]
     if options.debug:
@@ -394,7 +418,7 @@ if platform.supports_ninja_browse():
 # Search for generated headers relative to build dir.
 cflags.append('-I.')
 
-def shell_escape(str):
+def shell_escape(str: str) -> str:
     """Escape str such that it's interpreted as a single argument by
     the shell."""
 
@@ -415,6 +439,7 @@ n.variable('cflags', ' '.join(shell_escape(flag) for flag in cflags))
 if 'LDFLAGS' in configure_env:
     ldflags.append(configure_env['LDFLAGS'])
 n.variable('ldflags', ' '.join(shell_escape(flag) for flag in ldflags))
+
 n.newline()
 
 if platform.is_msvc():
@@ -471,48 +496,64 @@ if platform.supports_ninja_browse():
     n.newline()
 
 n.comment('the depfile parser and ninja lexers are generated using re2c.')
-def has_re2c():
+def has_re2c() -> bool:
     try:
         proc = subprocess.Popen(['re2c', '-V'], stdout=subprocess.PIPE)
-        return int(proc.communicate()[0], 10) >= 1103
+        return int(proc.communicate()[0], 10) >= 1503
     except OSError:
         return False
 if has_re2c():
     n.rule('re2c',
-           command='re2c -b -i --no-generation-date -o $out $in',
+           command='re2c -b -i --no-generation-date --no-version -o $out $in',
            description='RE2C $out')
     # Generate the .cc files in the source directory so we can check them in.
     n.build(src('depfile_parser.cc'), 're2c', src('depfile_parser.in.cc'))
     n.build(src('lexer.cc'), 're2c', src('lexer.in.cc'))
 else:
-    print("warning: A compatible version of re2c (>= 0.11.3) was not found; "
+    print("warning: A compatible version of re2c (>= 0.15.3) was not found; "
            "changes to src/*.in.cc will not affect your build.")
 n.newline()
 
-n.comment('Core source files all build into ninja library.')
 cxxvariables = []
 if platform.is_msvc():
     cxxvariables = [('pdb', 'ninja.pdb')]
+
+n.comment('Generate a library for `ninja-re2c`.')
+re2c_objs = []
+for name in ['depfile_parser', 'lexer']:
+    re2c_objs += cxx(name, variables=cxxvariables)
+if platform.is_msvc():
+    n.build(built('ninja-re2c.lib'), 'ar', re2c_objs)
+else:
+    n.build(built('libninja-re2c.a'), 'ar', re2c_objs)
+n.newline()
+
+n.comment('Core source files all build into ninja library.')
+objs.extend(re2c_objs)
 for name in ['build',
              'build_log',
              'clean',
              'clparser',
              'debug_flags',
-             'depfile_parser',
              'deps_log',
              'disk_interface',
              'dyndep',
              'dyndep_parser',
              'edit_distance',
+             'elide_middle',
              'eval_env',
              'graph',
              'graphviz',
-             'lexer',
+             'jobserver',
+             'json',
              'line_printer',
              'manifest_parser',
              'metrics',
+             'missing_deps',
              'parser',
+             'real_command_runner',
              'state',
+             'status_printer',
              'string_piece_util',
              'util',
              'version']:
@@ -520,6 +561,7 @@ for name in ['build',
 if platform.is_windows():
     for name in ['subprocess-win32',
                  'includes_normalize-win32',
+                 'jobserver-win32',
                  'msvc_helper-win32',
                  'msvc_helper_main-win32']:
         objs += cxx(name, variables=cxxvariables)
@@ -527,7 +569,9 @@ if platform.is_windows():
         objs += cxx('minidump-win32', variables=cxxvariables)
     objs += cc('getopt')
 else:
-    objs += cxx('subprocess-posix')
+    for name in ['jobserver-posix',
+                 'subprocess-posix']:
+        objs += cxx(name, variables=cxxvariables)
 if platform.is_aix():
     objs += cc('getopt')
 if platform.is_msvc():
@@ -559,40 +603,85 @@ if options.bootstrap:
     # build.ninja file.
     n = ninja_writer
 
-n.comment('Tests all build into ninja_test executable.')
+# Build the ninja_test executable only if the GTest source directory
+# is provided explicitly. Either from the environment with GTEST_SOURCE_DIR
+# or with the --gtest-source-dir command-line option.
+#
+# Do not try to look for an installed binary version, and link against it
+# because doing so properly is platform-specific (use the CMake build for
+# this).
+if options.gtest_source_dir:
+    gtest_src_dir = options.gtest_source_dir
+else:
+    gtest_src_dir = os.environ.get('GTEST_SOURCE_DIR')
 
-objs = []
-if platform.is_msvc():
-    cxxvariables = [('pdb', 'ninja_test.pdb')]
+if gtest_src_dir:
+    # Verify GoogleTest source directory, and add its include directory
+    # to the global include search path (even for non-test sources) to
+    # keep the build plan generation simple.
+    gtest_all_cc = os.path.join(gtest_src_dir, 'googletest', 'src', 'gtest-all.cc')
+    if not os.path.exists(gtest_all_cc):
+        print('ERROR: Missing GoogleTest source file: %s' % gtest_all_cc)
+        sys.exit(1)
 
-for name in ['build_log_test',
-             'build_test',
-             'clean_test',
-             'clparser_test',
-             'depfile_parser_test',
-             'deps_log_test',
-             'dyndep_parser_test',
-             'disk_interface_test',
-             'edit_distance_test',
-             'graph_test',
-             'lexer_test',
-             'manifest_parser_test',
-             'ninja_test',
-             'state_test',
-             'string_piece_util_test',
-             'subprocess_test',
-             'test',
-             'util_test']:
-    objs += cxx(name, variables=cxxvariables)
-if platform.is_windows():
-    for name in ['includes_normalize_test', 'msvc_helper_test']:
-        objs += cxx(name, variables=cxxvariables)
+    n.comment('Tests all build into ninja_test executable.')
 
-ninja_test = n.build(binary('ninja_test'), 'link', objs, implicit=ninja_lib,
-                     variables=[('libs', libs)])
-n.newline()
-all_targets += ninja_test
+    # Test-specific version of cflags, must include the GoogleTest
+    # include directory.
+    test_cflags = cflags.copy()
+    test_cflags.append('-I' + os.path.join(gtest_src_dir, 'googletest', 'include'))
 
+    test_variables = [('cflags', test_cflags)]
+    if platform.is_msvc():
+        test_variables += [('pdb', 'ninja_test.pdb')]
+
+    test_names = [
+        'build_log_test',
+        'build_test',
+        'clean_test',
+        'clparser_test',
+        'depfile_parser_test',
+        'deps_log_test',
+        'disk_interface_test',
+        'dyndep_parser_test',
+        'edit_distance_test',
+        'elide_middle_test',
+        'explanations_test',
+        'graph_test',
+        'jobserver_test',
+        'json_test',
+        'lexer_test',
+        'manifest_parser_test',
+        'ninja_test',
+        'state_test',
+        'string_piece_util_test',
+        'subprocess_test',
+        'test',
+        'util_test',
+    ]
+    if platform.is_windows():
+        test_names += [
+            'includes_normalize_test',
+            'msvc_helper_test',
+        ]
+
+    objs = []
+    for name in test_names:
+        objs += cxx(name, variables=test_variables)
+
+    # Build GTest as a monolithic source file.
+    # This requires one extra include search path, so replace the
+    # value of 'cflags' in our list.
+    gtest_all_variables = test_variables[1:] + [
+      ('cflags', test_cflags + ['-I' + os.path.join(gtest_src_dir, 'googletest') ]),
+    ]
+    # Do not use cxx() directly to ensure the object file is under $builddir.
+    objs += n.build(built('gtest_all' + objext), 'cxx', gtest_all_cc, variables=gtest_all_variables)
+
+    ninja_test = n.build(binary('ninja_test'), 'link', objs, implicit=ninja_lib,
+                         variables=[('libs', libs)])
+    n.newline()
+    all_targets += ninja_test
 
 n.comment('Ancillary executables.')
 
@@ -603,6 +692,7 @@ if platform.is_aix() and '-maix64' not in ldflags:
 
 for name in ['build_log_perftest',
              'canon_perftest',
+             'elide_middle_perftest',
              'depfile_parser_perftest',
              'hash_collision_bench',
              'manifest_parser_perftest',
@@ -684,7 +774,7 @@ if host.is_linux():
 
 n.build('all', 'phony', all_targets)
 
-n.close()
+n.close()  # type: ignore # Item "Bootstrap" of "Writer | Bootstrap" has no attribute "close"
 print('wrote %s.' % BUILD_FILENAME)
 
 if options.bootstrap:
