@@ -28,6 +28,7 @@
 #include <sys/time.h>
 #endif
 
+#include "elide_middle.h"
 #include "util.h"
 
 using namespace std;
@@ -46,10 +47,6 @@ LinePrinter::LinePrinter() : have_blank_line_(true), console_locked_(false) {
   }
 #endif
   supports_color_ = smart_terminal_;
-  if (!supports_color_) {
-    const char* clicolor_force = getenv("CLICOLOR_FORCE");
-    supports_color_ = clicolor_force && string(clicolor_force) != "0";
-  }
 #ifdef _WIN32
   // Try enabling ANSI escape sequence support on Windows 10 terminals.
   if (supports_color_) {
@@ -61,6 +58,10 @@ LinePrinter::LinePrinter() : have_blank_line_(true), console_locked_(false) {
     }
   }
 #endif
+  if (!supports_color_) {
+    const char* clicolor_force = getenv("CLICOLOR_FORCE");
+    supports_color_ = clicolor_force && std::string(clicolor_force) != "0";
+  }
 }
 
 void LinePrinter::Print(string to_print, LineType type) {
@@ -81,7 +82,7 @@ void LinePrinter::Print(string to_print, LineType type) {
     CONSOLE_SCREEN_BUFFER_INFO csbi;
     GetConsoleScreenBufferInfo(console_, &csbi);
 
-    to_print = ElideMiddle(to_print, static_cast<size_t>(csbi.dwSize.X));
+    ElideMiddleInPlace(to_print, static_cast<size_t>(csbi.dwSize.X));
     if (supports_color_) {  // this means ENABLE_VIRTUAL_TERMINAL_PROCESSING
                             // succeeded
       printf("%s\x1B[K", to_print.c_str());  // Clear to end of line.
@@ -108,7 +109,7 @@ void LinePrinter::Print(string to_print, LineType type) {
     // line-wrapping.
     winsize size;
     if ((ioctl(STDOUT_FILENO, TIOCGWINSZ, &size) == 0) && size.ws_col) {
-      to_print = ElideMiddle(to_print, size.ws_col);
+      ElideMiddleInPlace(to_print, size.ws_col);
     }
     printf("%s", to_print.c_str());
     printf("\x1B[K");  // Clear to end of line.
@@ -118,6 +119,7 @@ void LinePrinter::Print(string to_print, LineType type) {
     have_blank_line_ = false;
   } else {
     printf("%s\n", to_print.c_str());
+    fflush(stdout);
   }
 }
 
